# -*- coding: utf-8 -*-

import sys
import argparse
import lamvery
import logging
from termcolor import colored
from lamvery.actions import (
    InitAction,
    ArchiveAction,
    ConfigureAction,
    DeployAction,
    DecryptAction,
    EncryptAction,
    EventsAction,
    InvokeAction,
    RollbackAction,
    SetAliasAction
)


def init(args):
    InitAction(args).action()


def archive(args):
    ArchiveAction(args).action()


def configure(args):
    ConfigureAction(args).action()


def deploy(args):
    DeployAction(args).action()


def decrypt(args):
    DecryptAction(args).action()


def encrypt(args):
    EncryptAction(args).action()


def events(args):
    EventsAction(args).action()


def invoke(args):
    InvokeAction(args).action()


def rollback(args):
    RollbackAction(args).action()


def set_alias(args):
    SetAliasAction(args).action()


def main():
    a_args = ('-a', '--alias',)
    a_kwargs = {
        'help': 'Alias for a version of the function',
        'default': None
    }
    c_args = ('-c', '--conf-file',)
    c_kwargs = {
        'help': 'Configuration YAML file (default: .lamvery.yml)',
        'default': '.lamvery.yml'
    }
    d_args = ('-d', '--dry-run',)
    d_kwargs = {
        'help': 'Dry run',
        'action': 'store_true',
        'default': False
    }
    k_args = ('-k', '--keep-empty-events',)
    k_kwargs = {
        'help': 'Keep the event rules that does not have any targets.',
        'action': 'store_true',
        'default': False
    }
    sf_args = ('-s', '--single-file',)
    sf_kwargs = {
        'help': 'Only use the main lambda function file',
        'action': 'store_true',
        'default': False
    }
    l_args = ('-l', '--no-libs',)
    l_kwargs = {
        'help': 'Archiving without all libraries',
        'action': 'store_true',
        'default': False
    }
    n_args = ('-n', '--secret-name',)
    n_kwargs = {
        'help': 'The name of the secret value',
        'default': None
    }
    p_args = ('-p', '--publish')
    p_kwargs = {
        'help': 'Publish the version as an atomic operation',
        'action': 'store_true',
        'default': False
    }
    s_args = ('-s', '--store',)
    s_kwargs = {
        'help': 'Store encripted value to configuration file (default: .lamvery.yml)',
        'action': 'store_true',
        'default': False
    }
    v_args = ('-v', '--version',)
    v_kwargs = {
        'help': 'Version of the function',
        'default': None
    }

    parser = argparse.ArgumentParser(
        description='Yet another deploy tool for AWS Lambda in the virtualenv environment.',
        epilog='Lamvery version: {}'.format(lamvery.__version__))
    subparsers = parser.add_subparsers(title='subcommands')

    init_parser = subparsers.add_parser(
        'init',
        help='Generate initial configuration file')
    init_parser.add_argument(*c_args, **c_kwargs)
    init_parser.set_defaults(func=init)

    archive_parser = subparsers.add_parser(
        'archive',
        help='Archive your code and libraries to <your-function-name>.zip')
    archive_parser.add_argument(*c_args, **c_kwargs)
    archive_parser.add_argument(*sf_args, **sf_kwargs)
    archive_parser.add_argument(*l_args, **l_kwargs)
    archive_parser.set_defaults(func=archive)

    set_alias_parser = subparsers.add_parser(
        'set-alias',
        help='Set alias to a version of the function')
    set_alias_parser.add_argument(*a_args, **a_kwargs)
    set_alias_parser.add_argument(*c_args, **c_kwargs)
    set_alias_parser.add_argument(*d_args, **d_kwargs)
    set_alias_parser.add_argument(*v_args, **v_kwargs)
    set_alias_parser.set_defaults(func=set_alias)

    configure_parser = subparsers.add_parser(
        'configure',
        help='Update the remote configuration')
    configure_parser.add_argument(*c_args, **c_kwargs)
    configure_parser.add_argument(*d_args, **d_kwargs)
    configure_parser.set_defaults(func=configure)

    deploy_parser = subparsers.add_parser(
        'deploy',
        help='Deploy your code and libraries,' +
             'Update the remote configuration, Set alias (optional)')
    deploy_parser.add_argument(*a_args, **a_kwargs)
    deploy_parser.add_argument(*c_args, **c_kwargs)
    deploy_parser.add_argument(*d_args, **d_kwargs)
    deploy_parser.add_argument(*sf_args, **sf_kwargs)
    deploy_parser.add_argument(*l_args, **l_kwargs)
    deploy_parser.add_argument(*p_args, **p_kwargs)
    deploy_parser.set_defaults(func=deploy)

    encrypt_parser = subparsers.add_parser('encrypt', help='Encrypt a text value using KMS')
    encrypt_parser.add_argument('text', help='The text to be encrypted')
    encrypt_parser.add_argument(*c_args, **c_kwargs)
    encrypt_parser.add_argument(*n_args, **n_kwargs)
    encrypt_parser.add_argument(*s_args, **s_kwargs)
    encrypt_parser.set_defaults(func=encrypt)

    decrypt_parser = subparsers.add_parser('decrypt', help='Decrypt the secret value using KMS')
    decrypt_parser.add_argument(*c_args, **c_kwargs)
    decrypt_parser.add_argument(*n_args, **n_kwargs)
    decrypt_parser.set_defaults(func=decrypt)

    events_parser = subparsers.add_parser(
        'events',
        help='Configure all events of CloudWatchEvents using the function')
    events_parser.add_argument(*c_args, **c_kwargs)
    events_parser.add_argument(*d_args, **d_kwargs)
    events_parser.add_argument(*k_args, **k_kwargs)
    events_parser.set_defaults(func=events)

    invoke_parser = subparsers.add_parser(
        'invoke',
        help='Invoke the function')
    invoke_parser.add_argument(
        'json', default='{}', help='The JSON string or file that pass to the function')
    invoke_parser.add_argument(*a_args, **a_kwargs)
    invoke_parser.add_argument(*c_args, **c_kwargs)
    invoke_parser.add_argument(*v_args, **v_kwargs)
    invoke_parser.set_defaults(func=invoke)

    rollback_parser = subparsers.add_parser(
        'rollback',
        help='Rollback your code and libraries')
    rollback_parser.add_argument(*a_args, **a_kwargs)
    rollback_parser.add_argument(*c_args, **c_kwargs)
    rollback_parser.add_argument(*v_args, **v_kwargs)
    rollback_parser.set_defaults(func=rollback)

    try:
        args = parser.parse_args()
        args.func(args)
        sys.exit(0)
    except Exception as e:
        msg = str(e)
        logging.exception(msg)
        sys.exit(colored(msg, 'red'))



# -*- coding: utf-8 -*-

from lamvery.actions.base import BaseAction


class EncryptAction(BaseAction):

    def __init__(self, args):
        super(EncryptAction, self).__init__(args)
        self._text = args.text
        self._name = args.secret_name
        self._store = args.store

    def action(self):
        cipher_text = self.get_client().encrypt(
            self._config.get_secret().get('key_id'), self._text)

        if self._store:
            self._config.store_secret(self._name, cipher_text)
        else:
            print(cipher_text)



# -*- coding: utf-8 -*-

from lamvery.actions.base import BaseAction


class DecryptAction(BaseAction):

    def __init__(self, args):
        super(DecryptAction, self).__init__(args)
        self._name = args.secret_name

    def action(self):
        text = self.get_client().decrypt(
            self._config.get_secret().get('cipher_texts').get(self._name))
        print(text)



# -*- coding: utf-8 -*-

from lamvery.actions.base import BaseAction


class SetAliasAction(BaseAction):

    def __init__(self, args):
        super(SetAliasAction, self).__init__(args)
        self._alias = args.alias
        if hasattr(args, 'version'):
            self._version = args.version
        else:
            self._version = None

    def action(self):
        alias_name = self.get_alias_name()
        version = self.get_version()
        func_name = self._config.get_function_name()
        client = self.get_client()

        if alias_name is None:
            raise Exception(
                'Please specify an alias by `-a` option or `default_alias` configuration.')

        current_alias = client.get_alias(func_name, alias_name)
        self._print_alias_diff(alias_name, current_alias, version)

        if len(current_alias) > 0:
            client.update_alias(func_name, alias_name, version)
        else:
            client.create_alias(func_name, alias_name, version)

    def _print_alias_diff(self, name, current, version):
        self._logger.warn(
            '[Alias] {name}: {cur} -> {new}'.format(
                name=name, cur=current.get('FunctionVersion'), new=version))

    def get_alias_name(self):
        if self._alias is not None:
            return self._alias
        return self._config.get_default_alias()

    def get_version(self):
        if self._version is None:
            return '$LATEST'
        return self._version



# -*- coding: utf-8 -*-

from lamvery.actions.base import BaseAction
from lamvery.actions.configure import CONF_DIFF_KEYS
from lamvery.actions.set_alias import SetAliasAction
from lamvery.archive import Archive
from lamvery.utils import previous_alias


class DeployAction(BaseAction):

    def __init__(self, args):
        super(DeployAction, self).__init__(args)
        self._publish = args.publish
        self._set_alias = SetAliasAction(args)
        self._single_file = args.single_file
        self._no_libs = args.no_libs

    def action(self):
        archive_name = self._config.get_archive_name()
        function_filename = self._config.get_function_filename()
        secret = self._config.generate_lambda_secret()
        exclude = self._config.get_exclude()
        archive = Archive(filename=archive_name,
                          function_filename=function_filename,
                          secret=secret,
                          single_file=self._single_file,
                          no_libs=self._no_libs,
                          exclude=exclude,
                          runtime=self._config.get_runtime())
        func_name = self._config.get_function_name()
        local_conf = self._config.get_configuration()
        zipfile = archive.create_zipfile()
        client = self.get_client()
        remote_conf = client.get_function_conf(func_name)
        alias_name = self._set_alias.get_alias_name()
        remote_size = client.calculate_capacity()
        local_size = archive.get_size()
        new_version = None
        cur_version = None

        if len(remote_conf) == 0:
            self._logger.warn(
                '[Function] Create new function "{}"'.format(func_name))

        self._print_diff(
            prefix='[Function]',
            remote=remote_conf, local=local_conf, keys=CONF_DIFF_KEYS)

        if len(remote_conf) > 0:

            if self._enable_versioning():
                cur_version = client.get_alias(
                    func_name, alias_name).get('FunctionVersion')
            else:
                local_size -= remote_conf['CodeSize']

            client.update_function_conf(local_conf)
            self._print_capacity(remote=remote_size, local=local_size)
            new_version = client.update_function_code(
                zipfile, local_conf, self._enable_versioning())

        else:
            if self._enable_versioning():
                local_size *= 2

            self._print_capacity(
                remote=remote_size, local=local_size)
            new_version = client.create_function(
                zipfile, local_conf, self._enable_versioning())

        zipfile.close()

        if new_version is not None:
            self._logger.info(
                '[Function] Deployed version: {}'.format(new_version))

        if cur_version is not None:
            self._logger.info(
                '[Function] Previous version: {}'.format(cur_version))
            self._set_alias._alias = previous_alias(alias_name)
            self._set_alias._version = cur_version
            self._set_alias.action()

        if alias_name is not None:
            self._set_alias._alias = alias_name
            self._set_alias._version = new_version
            self._set_alias.action()

    def _enable_versioning(self):
        if self._publish:
            return True
        return self._config.enable_versioning()

    def _print_capacity(self, remote, local):
        self._logger.warn(
            '[Function] Capacity: {r} Bytes -> {t} Bytes'.format(
                r='{:,d}'.format(remote), t='{:,d}'.format(remote + local)))



# -*- coding: utf-8 -*-

import json
from lamvery.actions.base import BaseAction

EVENT_RULE_DIFF_KEYS = [
    ('Description', 'description',),
    ('EventPattern', 'pattern',),
    ('RoleArn', 'role',),
    ('ScheduleExpression', 'schedule',),
    ('State', 'state',),
]

EVENT_TARGET_DIFF_KEYS = [
    ('Input', 'input',),
    ('InputPath', 'input_path',),
]


class EventsAction(BaseAction):

    def __init__(self, args):
        super(EventsAction, self).__init__(args)
        self._keep_empty = args.keep_empty_events

    def action(self):
        client = self.get_client()
        func_name = self._config.get_function_name()
        conf = client.get_function_conf(func_name)

        if len(conf) == 0:
            msg = '"{}" function is not exists. Please `deploy` at first.'.format(func_name)
            raise Exception(msg)

        arn = conf['FunctionArn']
        local_rules = self._config.get_events()
        remote_rules = client.get_rules_by_target(arn)

        self._clean(remote_rules, local_rules, arn, func_name)
        self._put_rules(remote_rules, local_rules, func_name)
        self._put_targets(local_rules, arn)

    def _put_rules(self, remote, local, function):
        client = self.get_client()

        for l in local:
            l['state'] = self._convert_state(l.get('disabled'))

            r = self._search_rule(remote, l['rule'])
            if len(r) == 0:
                self._logger.warn(
                    '[EventRule] Create new event rule "{}"'.format(l['rule']))

            self._print_diff(
                prefix='[EventRule] {}:'.format(l['rule']),
                keys=EVENT_RULE_DIFF_KEYS,
                remote=r, local=l)

            ret = client.put_rule(l)
            client.add_permission(function, l['rule'], ret.get('RuleArn'))

    def _convert_state(self, disabled):
        if disabled:
            return 'DISABLED'
        else:
            return 'ENABLED'

    def _search_rule(self, rules, name):
        for r in rules:
            if name in [r.get('Name'), r.get('rule')]:
                return r
        return {}

    def _exist_rule(self, rules, name):
        return len(self._search_rule(rules, name)) > 0

    def _put_targets(self, local, arn):
        client = self.get_client()

        for l in local:
            targets = client.get_targets_by_rule(l['rule'])

            for lt in l['targets']:
                if 'input' in lt:
                    lt['input'] = json.dumps(lt['input'])

                diff_r = {}
                for rt in targets:
                    if rt['Id'] == lt['id']:
                        diff_r = rt
                        break
                    self._logger.warn(
                        '[EventRule] {}: Add "{}" to targets'.format(l['rule'], lt['id']))

                self._print_diff(
                    prefix='[EventTarget] {}:'.format(lt['id']),
                    keys=EVENT_TARGET_DIFF_KEYS,
                    remote=diff_r, local=lt)

            client.put_targets(
                rule=l['rule'], targets=l['targets'], arn=arn)

    def _exist_target(self, targets, target_id):
        for t in targets:
            if target_id in [t.get('id'), t.get('Id')]:
                return True
        return False

    def _clean(self, remote, local, arn, function):
        client = self.get_client()

        for r in remote:
            targets = client.get_targets_by_rule(r['Name'])
            target_ids = []

            l = self._search_rule(local, r['Name'])

            for rt in targets:
                msg = '[EventRule] {}: Remove undifined event target "{}"'.format(
                    r['Name'], rt['Id'])
                if len(l) > 0:
                    if not self._exist_target(l['targets'], rt['Id']):
                        self._logger.warn(msg)
                        target_ids.append(rt['Id'])
                elif rt['Arn'] == arn:
                    self._logger.warn(msg)
                    target_ids.append(rt['Id'])

            if len(target_ids) > 0:
                client.remove_targets(r['Name'], target_ids)

            if len(targets) == len(target_ids) and not self._keep_empty:
                self._logger.warn(
                    '[EventRule] Delete the event rule "{}" that does not have any targets'.format(
                        r['Name']))
                client.delete_rule(r['Name'])
                client.remove_permission(function, r['Name'])



# -*- coding: utf-8 -*-

from lamvery.actions.base import BaseAction
from lamvery.actions.set_alias import SetAliasAction


class RollbackAction(BaseAction):

    def __init__(self, args):
        super(RollbackAction, self).__init__(args)
        self._set_alias = SetAliasAction(args)

    def action(self):
        func_name = self._config.get_function_name()
        client = self.get_client()
        remote_conf = client.get_function_conf(func_name)
        alias_name = self._set_alias.get_alias_name()

        if len(remote_conf) == 0:
            raise Exception(
                '"{}" function is not exists. Please `deploy` at first.'.format(func_name))

        pre_version = client.get_previous_version(func_name, alias_name)
        if pre_version is None:
            raise Exception(
                'There is no previous version. ' +
                'Please `deploy` with `publish` option or `versioning` configuration.')

        self._logger.info(
            '[Function] Previous version: {}'.format(pre_version))

        self._set_alias._version = pre_version
        self._set_alias.action()



# -*- coding: utf-8 -*-

import yaml
import os
import re

from collections import OrderedDict
from jinja2 import Environment, FileSystemLoader

RUNTIME_PY_27 = 'python2.7'
RUNTIME_NODE_JS = 'nodejs'

RUNTIME_AND_EXTENSION = {
    RUNTIME_PY_27: 'py',
    RUNTIME_NODE_JS: 'js'
}


def represent_odict(dumper, instance):
    return dumper.represent_mapping(u'tag:yaml.org,2002:map', instance.items())

yaml.add_representer(OrderedDict, represent_odict)
yaml.add_constructor(
    yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
    lambda loader, node: OrderedDict(loader.construct_pairs(node)))


class Config:

    def __init__(self, conf_file):
        self._file = conf_file
        self._template_env = Environment(loader=FileSystemLoader('./', encoding='utf8'))

    def load(self, file):
        tmpl = self._template_env.get_template(file)
        return yaml.load(tmpl.render({'env': os.environ}))

    def load_conf(self):
        return self.load(self._file)

    def load_events(self):
        return self.load(self.get_event_file())

    def get_event_file(self):
        return self.load_conf().get('event_file', '.lamvery.event.yml')

    def load_secret(self):
        return self.load(self.get_secret_file())

    def get_secret_file(self):
        return self.load_conf().get('secret_file', '.lamvery.secret.yml')

    def load_exclude(self):
        return self.load(self.get_exclude_file())

    def get_exclude_file(self):
        return self.load_conf().get('exclude_file', '.lamvery.exclude.yml')

    def load_raw_secret(self):
        txt = open(self.get_secret_file(), 'r').read()
        return yaml.load(self.escape(txt))

    def escape(self, txt):
        txt = txt.replace("'", "''")
        txt = txt.replace(
            self._template_env.variable_start_string,
            "'" + self._template_env.variable_start_string)
        txt = txt.replace(
            self._template_env.block_start_string,
            "'" + self._template_env.block_start_string)
        txt = txt.replace(
            self._template_env.variable_end_string,
            self._template_env.variable_end_string + "'")
        txt = txt.replace(
            self._template_env.block_end_string,
            self._template_env.block_end_string + "'")
        return txt

    def unescape(self, txt):
        txt = txt.replace("''", "'")
        txt = txt.replace(
            "'" + self._template_env.variable_start_string,
            self._template_env.variable_start_string)
        txt = txt.replace(
            "'" + self._template_env.block_start_string,
            self._template_env.block_start_string)
        txt = txt.replace(
            self._template_env.variable_end_string + "'",
            self._template_env.variable_end_string)
        txt = txt.replace(
            self._template_env.block_end_string + "'",
            self._template_env.block_end_string)
        return txt

    def get_configuration(self):
        return self.load_conf().get('configuration')

    def get_secret(self):
        return self.load_secret()

    def get_events(self):
        events = self.load_events()
        if events is None:
            return []
        return events

    def get_default_alias(self):
        return self.load_conf().get('default_alias')

    def enable_versioning(self):
        return bool(self.load_conf().get('versioning'))

    def generate_lambda_secret(self):
        return {
            'region': self.get_region(),
            'cipher_texts': self.get_secret().get('cipher_texts')
        }

    def get_function_name(self):
        if os.path.exists(self._file):
            return self.get_configuration().get('name')
        else:
            return os.path.basename(os.getcwd())

    def get_function_filename(self):
        ext = RUNTIME_AND_EXTENSION.get(self.get_runtime(), "py")
        return '{}.{}'.format(
            self.get_configuration().get('handler').split('.')[0], ext)

    def get_runtime(self):
        return self.get_configuration().get('runtime')

    def get_archive_name(self):
        return '{}.zip'.format(self.get_function_name())

    def get_region(self):
        if os.path.exists(self._file):
            return self.load_conf().get('region')
        else:
            return None

    def get_profile(self):
        return self.load_conf().get('profile')

    def get_exclude(self):
        exclude = self.load_exclude()
        if exclude is None:
            return []
        return exclude

    def get_default(self):
        init_config = OrderedDict()
        init_config['name'] = self.get_function_name()
        init_config['runtime'] = RUNTIME_PY_27
        init_config['role'] = 'arn:aws:iam::<account-number>:role/<role>'
        init_config['handler'] = 'lambda_function.lambda_handler'
        init_config['description'] = 'This is a sample lambda function.'
        init_config['timeout'] = 10
        init_config['memory_size'] = 128

        init_yaml = OrderedDict()
        init_yaml['profile'] = None
        init_yaml['region'] = 'us-east-1'
        init_yaml['versioning'] = False
        init_yaml['default_alias'] = None
        init_yaml['configuration'] = init_config

        return init_yaml

    def get_default_events(self):
        targets = [
            OrderedDict([
                ('id', '<unique-target-id>',),
                ('input', {'this': [{'is': 'a'}, {'sample': 'input'}]},),
                ('input_path', 'json.path.format',)])]
        event = OrderedDict()
        event['rule'] = 'sample-rule-name'
        event['description'] = 'This is a sample CloudWatchEvent'
        event['schedule'] = 'rate(5 minutes)'
        event['targets'] = targets
        init_events = [event]

        return init_events

    def get_default_secret(self):
        init_secret = OrderedDict()
        init_secret['key_id'] = '<key-id>'
        init_secret['cipher_texts'] = OrderedDict()

        return init_secret

    def get_default_exclude(self):
        return [
            '^{}$'.format(re.escape(self._file)),
            '^{}$'.format(re.escape(self.get_event_file())),
            '^{}$'.format(re.escape(self.get_secret_file())),
            '^{}$'.format(re.escape(self.get_exclude_file()))]

    def write(self, conf, path):
        txt = yaml.dump(
            conf,
            default_flow_style=False,
            allow_unicode=True)
        open(path, 'w').write(self.unescape(txt))

    def store_secret(self, key, value):
        conf = self.load_raw_secret()

        if 'cipher_texts' not in conf:
            conf['cipher_texts'] = {}

        conf['cipher_texts'][key] = value
        self.write(conf, self.get_secret_file())



# -*- coding: utf-8 -*-

import os
import sys
from termcolor import cprint
from lamvery.actions.base import BaseAction


class InitAction(BaseAction):

    def __init__(self, args):
        super(InitAction, self).__init__(args)
        self._conf_file = args.conf_file

    def action(self):
        if self._needs_write(self._conf_file):
            self._config.write(self._config.get_default(), self._conf_file)
            self._logger.info(
                'Output initial file: {}'.format(self._conf_file))

        files = {
            self._config.get_event_file(): self._config.get_default_events(),
            self._config.get_secret_file(): self._config.get_default_secret(),
            self._config.get_exclude_file(): self._config.get_default_exclude(),
        }

        for f, c in files.items():
            if self._needs_write(f):
                self._config.write(c, f)
                self._logger.info(
                    'Output initial file: {}'.format(f))

    def _needs_write(self, path):
        ret = True
        if os.path.exists(path):
            cprint('Overwrite {}? [y/n]: '.format(path), 'yellow', file=sys.stderr, end="")
            y_n = sys.stdin.readline()
            if not y_n.startswith('y'):
                ret = False
        return ret
